var http = require('http');
var router = require('choreographer').router();
var fs = require('fs');
var pathlib = require('path');
var uri = require('url');
var log = require('./log');
var sys = require('sys');
var Script = process.binding('evals').Script;
var child_process = require('child_process');
var mime = require('./content-type');

exports.log_levels = log.levels;

var settings;

var server;

var root;

exports.start = function(_settings) {
  settings = _settings;
  log.level = settings.log_level;
  // Start the fileserver side
  start_fileserver( start_webserver );
}

router.get('/', function(req, res, path) {
  serve_static_file('static/index.html', req, res);
});

router.get('/static/**', function(req, res, path) {
  serve_static_file('static/'+path, req, res);
});

router.get('/api/browse', function(req, res) {
  try {
    res.writeHead(200, {'Content-Type': 'text/plain'});
    res.write(JSON.stringify(root));
  }
  catch (e) {
    res.writeHead(500, {'Content-Type': 'text/plain'});
    res.write(e.toString());
  }
  res.end();
});

router.get('/api/launch', function(req,res) {
  var query = uri.parse(req.url, true).query;
  var name = query['name'];
  var command = settings.command_launch + ' ' + name;
  res.writeHead(200, {'Content-Type': 'text/plain'});
  res.write('Executing:\r\n');
  res.write(command+'\r\n');
  run_process(command, res);
  res.write('PID:\r\n');
  res.write(process.pid+'\r\n');
});

router.get('/api/list', function(req,res) {
  var command = 'ps -A | grep '+settings.command_grep;
  res.writeHead(200, {'Content-Type': 'text/plain'});
  var process = child_process.exec(command, function(err, stdout, stderr) {
    var out = [];
    var lines = stdout.split('\n');
    for (var i in lines) {
      var matches = lines[i].match(/(^\d+).*:..\... (.*)/);
      if (matches) {
        if (matches[2].indexOf('grep '+settings.command_grep) > -1) continue;
        out.push({
          'pid' : matches[1],
          'name': matches[2],
          'kill':'/api/kill/'+matches[1]
        });
      }
    }
    res.write(JSON.stringify(out));
    res.end();
  });
});

router.get('/api/kill/*', function(req,res,pid) {
  res.write('killing pid '+pid+'\r\n');
  run_process('kill '+pid, res);
})

function start_fileserver(next) {
  log.info('Starting fileserver');
  root = read_dir(settings.data);
  if (next) next();
}

function read_dir(fullname,subname) {
  if (!subname) subname = fullname;
  var out = { 
    name: subname,
    files: [],
    dirs:  []
  };
  // 
  var data = fs.readdirSync(fullname);
  for (var n=0;n<data.length;n++) {
    var subname = fullname + '/' + data[n];
    var stats = fs.statSync(subname);
    if (stats.isFile()) {
      out.files.push({ 
        'name':data[n],
        'fullname':subname,
        'launch':'/api/launch?name='+subname
      });
    }
    else {
      out.dirs.push(read_dir(subname, data[n]));
    }
  }
  return out;
}


function start_webserver() {
  // Starting server 
  log.debug("settings = " + JSON.stringify(settings));
  log.info( "Starting webserver on port", settings.port);
  // Handle requests
  server = http.createServer(router);
  server.listen(settings.port);
  // Handle connection closing
  server.addListener('connection', function(connection) {
    connection.setTimeout(settings.timeout_milliseconds);
    connection.addListener('timeout', function() {
      log.debug("Connection from",connection.remoteAddress,"timed out. Closing it...");
      connection.destroy();
    });
  });
};

function run_process(command, res) {
  child_process.exec(command, function(err, stdout, stderr) {
    res.write('stdout:\r\n' + stdout+'\r\n');
    res.write('stderr:\r\n' + stderr+'\r\n');
    res.write('err:\r\n' + err+'\r\n');
    res.end();
  });
}


function serve_static_file(path, req, resp) {
    fs.stat(path, function (err, stats) {
        if (err) {
            return file_not_found();
        }
        if (!stats.isFile()) 
            return file_not_found();
        var if_modified_since = req.headers['if-modified-since'];

        try {
            var readStream = fs.createReadStream(path);
        } 
        catch (err) {
            log.debug("fs.createReadStream(",path,") error: ",sys.inspect(err,true));
            return file_not_found();
        }

        resp.writeHead(200, {
            'Content-Length':stats.size, 
            'Content-Type':mime.mime_type(path),
            'Last-Modified':stats.mtime
        });
        sys.pump(readStream, resp, function() {
            log.debug('pumped',path);
        });

        req.connection.addListener('timeout', function() {
            /* dont destroy it when the fd's already closed */
            if (readStream.readable) {
                log.debug('timed out. destroying file read stream');
                readStream.destroy();
            }
        });

        readStream.addListener('fd', function(fd) {
            log.debug("opened",path,"on fd",fd);
        });

        readStream.addListener('error', function (err) {
            log.error('error reading',path,sys.inspect(err));
            resp.end('');
        });
        resp.addListener('error', function (err) {
            log.error('error writing',path,sys.inspect(err));
            readStream.destroy();
        });
    });


    function file_not_found() {
        log.debug("404 opening path: '"+path+"'");
        var body = "404: " + req.url + " not found.\n";
        resp.writeHead(404,{ 
            'Content-Length':body.length,
            'Content-Type':"text/plain"
        });
        if (req.method != 'HEAD') {
            resp.end(body, 'utf-8');
        } else {
            resp.end('');
        }
    }
}

exports.stop = function(callback) {
  if (server) {
    if (callback) server.addListener('close', callback);
    server.close();
  }
};


